/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 David Zeuthen <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <glib.h>
#include <polkit/polkit.h>
#include <polkit-dbus/polkit-dbus.h>

static void 
exit_not_privileged (PolKitResult pk_result, const char *pk_action)
{
        const char *pk_result_as_str;
        pk_result_as_str = polkit_result_to_string_representation (pk_result);
        fprintf (stderr, "%s %s", pk_result_as_str, pk_action);
        exit(1);
}

static gboolean
check_privileged (pid_t caller_pid, 
                  const char *pk_action, 
                  PolKitResult *pk_result)
{
        DBusError error;
        DBusConnection *bus;
        PolKitCaller *caller;
        PolKitContext *pol_ctx;
        PolKitAction *action;

        dbus_error_init (&error);
        bus = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
        if (bus == NULL) {
                fprintf (stderr, "cannot connect to system bus: %s: %s\n", error.name, error.message);
                dbus_error_free (&error);
                goto out;
        }

        action = polkit_action_new ();
        polkit_action_set_action_id (action, pk_action);

        caller = polkit_caller_new_from_pid (bus, caller_pid, &error);
        if (caller == NULL) {
                fprintf (stderr, "cannot get caller from dbus name\n");
                goto out;
        }

        pol_ctx = polkit_context_new ();
        if (!polkit_context_init (pol_ctx, NULL)) {
                fprintf (stderr, "cannot init polkit\n");
                goto out;
        }

        *pk_result = polkit_context_can_caller_do_action (pol_ctx, action, caller);
        return TRUE;
out:
        return FALSE;
}

static void
usage (void)
{
        fprintf (stderr, "usage: polkit-gnome-example-helper --frobnicate | --tweak\n");
        exit (2);
}

int
main (int argc, char *argv[])
{
        pid_t parent_pid;
        const char *pk_action;
        gboolean is_frobnicate = FALSE;
        gboolean is_tweak = FALSE;
        gboolean is_twiddle = FALSE;
        gboolean is_punch = FALSE;
        gboolean is_kick = FALSE;
        gboolean is_poke = FALSE;
        PolKitResult pk_result = POLKIT_RESULT_NO;

        printf ("In polkit-gnome-example-helper\n");


        if (argc < 2) {
                usage ();
        }

        if (strcmp (argv[1], "--frobnicate") == 0 && argc == 2) {
                is_frobnicate = TRUE;
        } else if (strcmp (argv[1], "--tweak") == 0 && argc == 2) {
                is_tweak = TRUE;
        } else if (strcmp (argv[1], "--twiddle") == 0 && argc == 2) {
                is_twiddle = TRUE;
        } else if (strcmp (argv[1], "--punch") == 0 && argc == 2) {
                is_punch = TRUE;
        } else {
                usage ();
        }

        parent_pid = getppid ();

        pk_action = NULL;
        if (is_frobnicate) {
                pk_action = "polkit-gnome-examples-frobnicate";
        } else if (is_tweak) {
                pk_action = "polkit-gnome-examples-tweak";
        } else if (is_twiddle) {
                pk_action = "polkit-gnome-examples-twiddle";
        } else if (is_punch) {
                pk_action = "polkit-gnome-examples-punch";
        } else if (is_kick) {
                pk_action = "polkit-gnome-examples-kick";
        } else if (is_poke) {
                pk_action = "polkit-gnome-examples-poke";
        } else {
                usage ();
        }

        /* check if we're privileged */
        if (!check_privileged (parent_pid, pk_action, &pk_result)) {
                /* PolicyKit don't know - just return 'NO' for now */
                printf ("we're not privileged: PK inquiry failed\n");
                exit_not_privileged (POLKIT_RESULT_NO, pk_action);
        }

        if (pk_result != POLKIT_RESULT_YES) {
                /* Caller is not privileged */
                printf ("we're not privileged: pk_result='%s'\n", polkit_result_to_string_representation (pk_result));
                exit_not_privileged (pk_result, pk_action);
        }

        /* caller is indeed privileged; carry out the deed.. */
        printf ("we're privileged - yeay!\n");

        /* TODO: now either: "frobnicate", "tweak" or "twiddle", "punch", "kick", "poke" */

        return 0;
}
