/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 David Zeuthen <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <glib-object.h>

#include <gtk/gtk.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <polkit/polkit.h>

static DBusGProxy *polkit_gnome_proxy;

static DBusGConnection *
get_session_bus (void)
{
        GError          *error;
        DBusGConnection *bus;
        DBusConnection  *connection;

        error = NULL;
        bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
        if (bus == NULL) {
                g_warning ("Couldn't connect to session bus: %s", error->message);
                g_error_free (error);
                goto out;
        }

        connection = dbus_g_connection_get_connection (bus);
 out:
        return bus;
}

static GtkWidget *button_frobnicate;
static GtkWidget *button_tweak;
static GtkWidget *button_twiddle;
static GtkWidget *button_punch;

static void
simple_button_cb (GtkButton *button, gpointer user_data)
{
        int status;
        int return_code;
        char *helper_stderr = NULL;
        GError *error = NULL;
        gboolean gained_privilege;
        const char *helper_cmdline;
        char *cmdline;
        char *message_markup;

        /* gross hack - but it works */
        if (button == GTK_BUTTON (button_frobnicate)) {
                helper_cmdline = "polkit-gnome-example-helper --frobnicate";
                message_markup = _("System policy prevents the Policy-gnome example helper from Frobnicating");
        } else if (button == GTK_BUTTON (button_tweak)) {
                helper_cmdline = "polkit-gnome-example-helper --tweak";
                message_markup = _("System policy prevents the Policy-gnome example helper from Tweaking");
        } else if (button == GTK_BUTTON (button_twiddle)) {
                helper_cmdline = "polkit-gnome-example-helper --twiddle";
                message_markup = _("System policy prevents the Policy-gnome example helper from Twiddling");
        } else if (button == GTK_BUTTON (button_punch)) {
                helper_cmdline = "polkit-gnome-example-helper --punch";
                message_markup = _("System policy prevents the Policy-gnome example helper from Punching");
        } else {
                g_warning ("Unknown button pressed");
                goto out;
        }

        cmdline = g_strdup_printf (LIBEXECDIR /* "." */ "/%s", helper_cmdline);
        g_debug ("launching '%s'", cmdline);
        g_spawn_command_line_sync (cmdline,
                                   NULL,
                                   &helper_stderr,
                                   &status,
                                   NULL);
        g_free (cmdline);

        return_code = WEXITSTATUS (status);
        g_debug ("helper return code was %d", return_code);

        if (return_code == 1) {
                char **tokens;
                PolKitResult polkit_result;
                char *polkit_result_textual;
                char *polkit_action;

                tokens = g_strsplit (helper_stderr, " ", 2);
                if (g_strv_length (tokens) != 2) {
                        g_warning ("helper return string '%s' malformed", helper_stderr);
                        g_strfreev (tokens);
                        goto out;
                }

                polkit_result_textual = tokens[0];
                polkit_action = tokens[1];

                if (!polkit_result_from_string_representation (polkit_result_textual, &polkit_result)) {
                        g_warning ("polkit_result_textual '%s' malformed", polkit_result_textual);
                        g_strfreev (tokens);
                        goto out;
                }

                g_debug ("helper refused; returned polkit_result='%s' (%d) and polkit_action='%s'", 
                         polkit_result_textual, polkit_result, polkit_action);

                /* now use PolicyKit-gnome to bring up an auth dialog */
                if (!dbus_g_proxy_call_with_timeout (polkit_gnome_proxy,
                                                     "ShowDialog",
                                                     INT_MAX,
                                                     &error,
                                                     /* parameters: */
                                                     G_TYPE_STRING, polkit_action,  /* action_id */
                                                     G_TYPE_STRING, message_markup, /* message_markup */
                                                     G_TYPE_STRING, "",             /* dbus_system_bus_name */
                                                     G_TYPE_INVALID,
                                                     /* return values: */
                                                     G_TYPE_BOOLEAN, &gained_privilege,
                                                     G_TYPE_INVALID)) {
                        g_warning ("Caught exception '%s'", error->message);
                        g_error_free (error);
                        g_strfreev (tokens);
                        goto out;
                }

                g_debug ("gained privilege = %d", gained_privilege);

                g_strfreev (tokens);
        }

out:
        g_free (helper_stderr);
}


int
main (int argc, char **argv)
{
        GMainLoop           *loop;
        DBusGConnection     *session_bus;
        int                  ret;

        ret = 1;

        if (!g_thread_supported ()) {
                g_thread_init (NULL);
        }
        dbus_g_thread_init ();
        g_type_init ();
        gtk_init (&argc, &argv);

        loop = g_main_loop_new (NULL, FALSE);


        session_bus = get_session_bus ();
        if (session_bus == NULL) {
                goto out;
        }

	polkit_gnome_proxy = dbus_g_proxy_new_for_name (session_bus,
                                                        "org.gnome.PolicyKit",           /* bus name */
                                                        "/org/gnome/PolicyKit/Manager",  /* object */
                                                        "org.gnome.PolicyKit.Manager");  /* interface */

        GtkWidget *window;
        GtkWidget *vbox;
        GtkWidget *label;

        window = gtk_window_new (GTK_WINDOW_TOPLEVEL);


        vbox = gtk_vbox_new (FALSE, 10);
        gtk_container_add (GTK_CONTAINER (window), vbox);

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Frobnicate</b> requires the user to authenticate."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_frobnicate = gtk_button_new_with_label (_("Frobnicate!"));
        g_signal_connect (button_frobnicate, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_frobnicate, FALSE, FALSE, 0);

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Tweak</b> requires a system administrator to authenticate."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_tweak = gtk_button_new_with_label (_("Tweak!"));
        g_signal_connect (button_tweak, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_tweak, FALSE, FALSE, 0);

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Twiddle</b> requires a system administrator to authenticate. Once authenticated, this privilege can be retained indefinitely."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_twiddle = gtk_button_new_with_label (_("Twiddle!"));
        g_signal_connect (button_twiddle, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_twiddle, FALSE, FALSE, 0);

        label = gtk_label_new (NULL);
        gtk_label_set_markup (GTK_LABEL (label), _("Making the helper <b>Punch</b> requires the user to authenticate. Once authenticated, this privilege can be retained for the remainder of the desktop session."));
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
        gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
        gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
        button_punch = gtk_button_new_with_label (_("Punch!"));
        g_signal_connect (button_punch, "clicked", G_CALLBACK (simple_button_cb), NULL);
        gtk_box_pack_start (GTK_BOX (vbox), button_punch, FALSE, FALSE, 0);

        gtk_window_set_title (GTK_WINDOW (window), _("PolicyKit-gnome demo"));
        g_signal_connect (window, "delete-event", exit, NULL);
        gtk_widget_show_all (window);

        g_main_loop_run (loop);

        g_main_loop_unref (loop);
        ret = 0;
out:
        return ret;
}
